/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2013-2016 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "setTimeStepViscAutoFunctionObject.H"
#include "turbulenceModel.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
namespace functionObjects
{
    defineTypeNameAndDebug(setTimeStepViscAutoFunctionObject, 0);

    addToRunTimeSelectionTable
    (
        functionObject,
        setTimeStepViscAutoFunctionObject,
        dictionary
    );
}
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::functionObjects::setTimeStepViscAutoFunctionObject::setTimeStepViscAutoFunctionObject
(
    const word& name,
    const Time& runTime,
    const dictionary& dict
)
:
    fvMeshFunctionObject(name, runTime, dict),
    time_(runTime),
    minViscTime_(0),
    currentTimeStep_(0),
    relaxationTimeFactor_(dict.lookupOrDefault<scalar>("relaxationTimeFactor", 1.0)),
    hasBeenExecuted_(false)
{
    read(dict);
}


// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

Foam::functionObjects::setTimeStepViscAutoFunctionObject::~setTimeStepViscAutoFunctionObject()
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

const Foam::Time&
Foam::functionObjects::setTimeStepViscAutoFunctionObject::time() const
{
    return time_;
}


bool Foam::functionObjects::setTimeStepViscAutoFunctionObject::adjustTimeStep()
{
    if(!hasBeenExecuted_) return false;

    scalar targetTimeStep =  min(
                                   timeStepPtr_().value(time_.timeOutputValue()),
                                   min( minViscTime_*viscDimLessTimePtr_().value(time_.timeOutputValue()),
                                   time_.deltaT().value())
                                );

//    Info << "time step before adjustment:  " << currentTimeStep_ << endl;
//    Info << "target time step:             " << targetTimeStep << endl;
    currentTimeStep_ = currentTimeStep_*(1.0-relaxationTimeFactor_)
                     + relaxationTimeFactor_*targetTimeStep;
    const_cast<Time&>(time()).setDeltaT
    (
        currentTimeStep_,
        false
    );

    return true;
}


bool Foam::functionObjects::setTimeStepViscAutoFunctionObject::read
(
    const dictionary& dict
)
{
    timeStepPtr_ = Function1<scalar>::New("deltaT", dict);
    viscDimLessTimePtr_ = Function1<scalar>::New("viscDimLessTime", dict);

    // Check that adjustTimeStep is active
    const dictionary& controlDict = time_.controlDict();

    Switch adjust;
    if
    (
       !controlDict.readIfPresent<Switch>("adjustTimeStep", adjust)
    || !adjust
    )
    {
        FatalIOErrorInFunction(dict)
            << "Need to set 'adjustTimeStep' true to allow timestep control"
            << exit(FatalIOError);
    }

    return true;
}


bool Foam::functionObjects::setTimeStepViscAutoFunctionObject::execute()
{

    volScalarField nuEff
    (
            mesh_.lookupObject<turbulenceModel>
            (
                turbulenceModel::propertiesName
            ).nuEff()
    );

    tmp<volScalarField> tVisc
    (
        new volScalarField
        (
                    IOobject
                    (
                        "tVisc",
                        mesh_.time().timeName(),
                        mesh_,
                        IOobject::NO_READ,
                        IOobject::NO_WRITE
                    ),
                    mesh_,
                    dimensionedScalar("0", dimTime, 0.0)
        )
    );
    tVisc.ref().ref() =  pow(mesh_.V(),0.666666666666666667) / nuEff;
    minViscTime_ = min( tVisc.ref().ref() ).value();

//    scalar maxViscosity = max(nuEff).value(); 
//    Info << "setTimeStepViscAuto: max. viscosity  = " << maxViscosity << endl;

    Info << "setTimeStepViscAuto: min. viscous time scale  = " << minViscTime_ << endl;

    if(!hasBeenExecuted_) //first call
        currentTimeStep_ = time_.deltaT().value();

    hasBeenExecuted_ = true;

    return true;
}


bool Foam::functionObjects::setTimeStepViscAutoFunctionObject::write()
{
    return true;
}


// ************************************************************************* //
